SystemOrganization addCategory: #DebuggerModel!

Object subclass: #DebugContext
	instanceVariableNames: 'context topContext method methodNode ranges'
	classVariableNames: 'MapCache MapCacheEntries'
	poolDictionaries: ''
	category: 'DebuggerModel'!

!DebugContext class methodsFor: 'instance creation' stamp: 'AndreiChis 1/7/2013 10:29'!
forContext: aContext 
	^ self new forContext: aContext ! !

!DebugContext methodsFor: 'evaluating actions' stamp: 'AndreiChis 1/7/2013 10:29'!
evaluate: expression
	^ Compiler new
		evaluate: expression
		in: context
		to: context receiver! !

!DebugContext methodsFor: 'initialization' stamp: 'AndreiChis 1/7/2013 10:29'!
forContext: aContext 
	method := aContext method.
	methodNode := method methodNode.

	ranges := Dictionary new.
	
	" Put all ranges with a possitive Key PC in the ranges dictionary "
	methodNode rawSourceRanges
		keysAndValuesDo: [ :node :range |
			node pc = 0
				ifFalse: [ ranges at: node pc put: range ]].
			
	context := aContext.
	topContext := aContext.

	" Sort all the Key PCs in the sortedKeyPCs "
	"sortedKeyPCs := SortedCollection new."
	" ranges keysDo: [ :pc | sortedKeyPCs add: pc ]."! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
gtDebuggerRetrieveVariableValuePairs
	|fieldList|

	fieldList := OrderedCollection new.
	fieldList 
		add: ('_self' ->  context receiver);
	      addAll: (context receiver class allInstVarNames collect: [:each | each -> (context receiver instVarNamed: each) ]).
	
	fieldList add: ('_thisContext' ->  context).
	 context == nil ifTrue: [^fieldList].
	fieldList 
		add: ('_stack top' -> ( context stackPtr > 0 ifTrue: [ context top])).
		
	(context tempNames size > 0) ifTrue: [
		fieldList addAll: ((1 to: context tempNames size) collect: [ :index || key value |
			key := context tempNames at: index.
			value := [context debuggerMap namedTempAt: index in: context] 
				on: Error 
				do: [value := (Text string: 'error obtaining field value' attribute: TextColor red) asString ].
			key -> value ] ) ].
	
	^ fieldList! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
locateClosureHome
	"In case the current context is a BlockContext locate the closureHome and ask the
	user to validate the new context. If closureHome is not found or the user does not 
	validate the new context, return nil. 
	If the current context is not a BlockContext return it."
	| closureHome result aText |
	
	context isExecutingBlock ifTrue: [
		closureHome := context activeHome.
		closureHome ifNil: [
			self inform: 'Method for block not found on stack, can''t edit and continue'.
		 	^ nil].
		 (self confirm: 'I will have to revert to the method from\which this block originated.  Is that OK?' withCRs) 
			ifFalse: [^ nil].
		^ closureHome].
	
	^ context! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
receiver
	^ context receiver! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
receiverClass
	"Answer the class of the receiver. 
	It may differ from 'self selectedClass' "
	
	^ context receiver class! !

!DebugContext methodsFor: 'evaluating actions' stamp: 'AndreiChis 1/7/2013 10:29'!
recompileCurrentMethodTo: aText 
	| classOfMethod selector |
	
	classOfMethod := self selectedClass.
	selector := classOfMethod parserClass new parseSelector: aText.
	(selector == self selectedMessageName or: [(self selectedMessageName beginsWith: 'DoIt')
		and: [selector numArgs = self selectedMessageName numArgs]]) ifFalse: [
			self inform: 'can''t change selector'. ^ nil].
		
	selector := classOfMethod 
		compile: aText 
		classified: self selectedMessageCategoryName 
		notifying: nil.
	selector isNil 
		ifFalse: [ ^ classOfMethod compiledMethodAt: selector. ]
		ifTrue: [ ^ nil ] 
		! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
selectedClass
	"Answer the class in which the current context's method was found."

	^ context methodClass! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
selectedMessageCategoryName
	"Answer the name of the message category of the message of the current context."

	^ self selectedClass organization categoryOfElement: self selectedMessageName! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
selectedMessageName
	"Answer the message selector of the current context.
	 If the method is unbound we can still usefully answer its old selector."

	| selector |
	selector := context methodSelector.
	^(selector ~~ context method selector
	    and: [selector beginsWith: 'DoIt'])
		ifTrue: [context method selector]
		ifFalse: [selector]! !

!DebugContext methodsFor: 'private' stamp: 'AndreiChis 1/7/2013 12:03'!
source
	"Answer the source code of the currently selected context."
	^ context sourceCode

	"Old implementation"
	"^ method holdsTempNames
		ifTrue: [ method getSource ]
		ifFalse: [ methodNode sourceText ]"! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
sourceRangeForPC: pc
	^ ranges at: pc ifAbsent: [ ^ 1 to: 0 ].! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 12:03'!
sourceText
	^ self source asText makeSelectorBold! !

!DebugContext methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
topContext: aContext
	topContext  := aContext ! !

Object subclass: #DebugSession
	instanceVariableNames: 'context process'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'DebuggerModel'!

!DebugSession class methodsFor: 'instance creation' stamp: 'AndreiChis 1/7/2013 10:29'!
process: aProcess context: aContext
	^ self new
		process: aProcess context: aContext! !

!DebugSession methodsFor: 'private process' stamp: 'AndreiChis 1/7/2013 10:29'!
activePC: aContext
	^ (self isLatestContext: aContext)
		ifTrue: [ context pc ]
		ifFalse: [ self previousPC: aContext ].! !

!DebugSession methodsFor: 'private' stamp: 'AndreiChis 1/7/2013 10:29'!
clear
	process := nil.
	context := nil! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
context
	^ context! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:40'!
createModelForContext: aContext
	^ (DebugContext forContext: aContext) topContext: context! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/7/2013 10:29'!
currentCodeRange: aContext
	^(self createModelForContext: aContext) sourceRangeForPC: (self activePC: aContext)! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/7/2013 10:29'!
filterTopContext: aContext
	|ctx|
	 
	ctx := aContext.
	(self shouldDisplayOnTopContext: ctx) ifTrue: [^ ctx].
	[(ctx := ctx sender) ~~ nil] whileTrue:
		[(self shouldDisplayOnTopContext: ctx) ifTrue: [^ ctx]].
	^ nil! !

!DebugSession methodsFor: 'testing' stamp: 'AndreiChis 1/7/2013 11:21'!
isInterruptedContextATest

	^ (self isTestObject: self context receiver) and: [ self isTestMethod: self context method of: self context receiver ]! !

!DebugSession methodsFor: 'private process' stamp: 'AndreiChis 1/7/2013 10:29'!
isLatestContext: aContext
	^ process suspendedContext == aContext! !

!DebugSession methodsFor: 'testing' stamp: 'AndreiChis 1/7/2013 11:17'!
isTestMethod: aCompiledMethod of: aTestCase

	^ aCompiledMethod selector = aTestCase selector! !

!DebugSession methodsFor: 'testing' stamp: 'AndreiChis 1/7/2013 11:17'!
isTestObject: anObject

	"I'm not sure this is the best way to doit because it creates a coupling with TestCase, but due that SUnit is part of the core I think it is not bad after all - Hernan'"
	^ anObject isKindOf: TestCase! !

!DebugSession methodsFor: 'private' stamp: 'AndreiChis 1/7/2013 11:21'!
prepareTestToRunAgain

	self context receiver prepareToRunAgain! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/7/2013 10:29'!
previousPC: aContext
	^ (aContext method pcPreviousTo: aContext pc)
		ifNil: [ aContext pc ]! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
process
	^ process! !

!DebugSession methodsFor: 'initialization' stamp: 'AndreiChis 1/7/2013 10:29'!
process: aProcess context: aContext
	aProcess stepToSendOrReturn.
	process := aProcess.
	context := self filterTopContext: aContext.
	context isNil ifTrue: [context := aContext].
	
! !

!DebugSession methodsFor: 'evaluating actions' stamp: 'AndreiChis 1/7/2013 10:29'!
recompileMethodTo: text inContext: recompilationContext
	| newMethod |				
				
	newMethod := (self createModelForContext: recompilationContext) recompileCurrentMethodTo: text.
	newMethod isNil 
		ifFalse: [ 
			self rewindContextToMethod: newMethod fromContext: recompilationContext. 
			^ newMethod ] 
		ifTrue: [
			^ nil ]! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 11:16'!
restart: selectedContext
	|ctx|
	ctx := process popTo: selectedContext.
	ctx == selectedContext ifTrue: [ 
		"Only restart the process if the stack was unwind"
		process
			restartTop;
			stepToSendOrReturn ].
	context := selectedContext.
	
	"Issue 3015 - Hernan"
	self isInterruptedContextATest ifTrue: [ self prepareTestToRunAgain ].
	
	self triggerEvent: #restart		
	! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 10:46'!
resume
	"Proceed execution of the receiver's model, starting after the expression at 
	which an interruption occurred."
	process resume.
	
	self triggerEvent: #resume! !

!DebugSession methodsFor: 'updating' stamp: 'AndreiChis 1/7/2013 10:29'!
rewindContextToMethod: aMethod fromContext: aContext
	"this method is typically to be used after a hot compilation of a method from the stack"
	| ctxt |
	ctxt :=  process popTo: aContext.
	
	ctxt == aContext
		ifFalse: [
			self inform: 'Method saved, but current context unchanged\because of unwind error. Click OK to see error' withCRs]
		ifTrue: [
			process
				restartTopWith: aMethod;
				stepToSendOrReturn ].
	
	self flag: 'is it realy necessary to handle quick methods separately?'.
	
	^ context := ctxt! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
selectedCodeRange: selectedContext
	^ self currentCodeRange: selectedContext! !

!DebugSession methodsFor: 'private context' stamp: 'AndreiChis 1/7/2013 10:29'!
shouldDisplayOnTopContext: aContext
	
	^ aContext method selector ~= #halt! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
stack
	^ context stack! !

!DebugSession methodsFor: 'accessing' stamp: 'AndreiChis 1/7/2013 10:29'!
stackOfSize: size
	^ context stackOfSize: size! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 10:29'!
stepInto
	self stepInto: context! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 11:00'!
stepInto: selectedContext
	process step: selectedContext.
	context := process stepToSendOrReturn.
	
	self triggerEvent: #stepInto! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 10:29'!
stepOver
	self stepOver: context! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 11:28'!
stepOver: selectedContext
	| newContext |
	newContext := process completeStep: selectedContext.
	context := 
		newContext == selectedContext
			ifTrue: [ process stepToSendOrReturn ]
			ifFalse: [ newContext ].

	self triggerEvent: #stepOver ! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 10:29'!
stepThrough
	self stepThrough: context! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 11:30'!
stepThrough: selectedContext
	"Send messages until you return to the present method context.
	 Used to step into a block in the method."
	
	process stepToHome: selectedContext.
	context := process stepToSendOrReturn.
			
	self triggerEvent: #stepThrough ! !

!DebugSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/7/2013 10:29'!
terminate
	self process isNil ifFalse: [self process terminate].
	self clear.
	
	self triggerEvent: #terminate! !

!DebugSession methodsFor: 'private' stamp: 'AndreiChis 1/7/2013 10:29'!
unwindTo: aContext
	|ctx|
	ctx := process popTo: aContext.
	ctx == aContext ifFalse: [ self error: 'Could not unwind stack' ].
	context := aContext! !

!DebugSession methodsFor: 'updating' stamp: 'AndreiChis 1/7/2013 10:29'!
updateWithContext: newContext fromProcess: aProcess
	self process: aProcess context: newContext
	 
	"self process: aProcess.
	self unwindTo: (self filterTopContextAndUnwindForContext: newContext).
	self process stepToSendOrReturn."! !
